const Clutter = imports.gi.Clutter;
const Gio = imports.gi.Gio;
const Lang = imports.lang;
const Main = imports.ui.main;
const Mainloop = imports.mainloop;
const PopupMenu = imports.ui.popupMenu;
const St = imports.gi.St;

const Me = imports.misc.extensionUtils.getCurrentExtension();
const DBus = Me.imports.dbus;

const LoopAll = "\u21BB";
const LoopOff = "\u2192";
const LoopTrack = "\u21A9";
const Next = "\u25B8\u25B8";
const Pause = "\u25AE\u25AE";
const Play = "\u25B6";
const Prev = "\u25C2\u25C2";
const ShuffleOff = "\u2192";
const ShuffleOn = "\u21AC";
const Stop = "\u25FC";

const MPRISPlayer = Lang.Class({
    Name: 'MPRISPlayer',
    Extends: PopupMenu.PopupMenuSection,
    
    _init: function (bus, owner) {
        this.parent();
        
        // DBus
        this._properties = DBus.Properties(bus);
        this._mediaplayer = DBus.MediaPlayer2(bus);
        this._mediaplayer_player = DBus.MediaPlayer2Player(bus);
        this._mediaplayer_playlists = DBus.MediaPlayer2Playlists(bus);
        this._mediaplayer_tracklist = DBus.MediaPlayer2TrackList(bus);
        
        // Name 
        this.name_item = new PopupMenu.PopupBaseMenuItem();
        this.name_item.connect("activate",Lang.bind(this,function () {
            if (this._mediaplayer.CanRaise)
                this._mediaplayer.RaiseRemote();
        }));
        this.name_box = new St.BoxLayout();
        this.name_icon = new St.Icon({icon_name:"audio-x-generic",icon_size:16,icon_type:St.IconType.SYMBOLIC,style_class:"icon"});
        this.name_label = new St.Label({text:this._mediaplayer.Identity});
        this.button_box = new St.BoxLayout();
        this.fulls_button = new St.Button({label:"\u2b06",style_class:"fulls"});
        this.fulls_button.connect("clicked",Lang.bind(this,function (){
            if (this._mediaplayer.CanSetFullscreen)
                this._mediaplayer.Fullscreen = !this._mediaplayer.Fullscreen;
        }));
        this.quit_button = new St.Button({label:"\u2718",style_class:"quit"});
        this.quit_button.connect("clicked",Lang.bind(this,function (){
            if (this._mediaplayer.CanQuit)
                this._mediaplayer.QuitRemote();
        }));
        this.name_box.add_actor(this.name_icon);
        this.name_box.add_actor(this.name_label);
        this.button_box.add_actor(this.fulls_button);
        this.button_box.add_actor(this.quit_button);
        this.name_item.addActor(this.name_box);
        this.name_item.addActor(this.button_box,{align:St.Align.END});
        this.addMenuItem(this.name_item);
        pln = bus.split(".")[3];
        if (pln=="rsfapemperor" || pln=="rsfapemperordev") this.setIcon("file:///usr/share/pixmaps/rsfap/emperor_icon.png");
        else if (pln=="rhythmbox") this.setIcon("file:///usr/share/pixmaps/rhythmbox-small.xpm");
        else if (pln=="banshee") this.setIcon("file:///usr/share/pixmaps/banshee.xpm");
        if (!this._mediaplayer.CanRaise) {
            this.name_item.setSensitive(false);
            this.name_item.actor.remove_style_pseudo_class('insensitive');
        }
        if (!this._mediaplayer.CanQuit)
            this.quit_button.set_style_pseudo_class("insensitive");
        if (!this._mediaplayer.CanSetFullscreen)
            this.fulls_button.set_style_pseudo_class("insensitive");
        else if (this._mediaplayer.Fullscreen)
            this.fulls_button.set_style_pseudo_class("active");
        
        // Info
        this.shuffle = false;
        this.loop = "None";
        this.trackid = "/org/mpris/MediaPlayer2/TrackList/NoTrack";
        this.info_item = new PopupMenu.PopupBaseMenuItem({reactive:false});
        this.info_box = new St.BoxLayout();
        this.meta_box = new St.BoxLayout({vertical:true});
        this.ctrl_box = new St.BoxLayout();
        this.info_art = new St.Icon({icon_name:"audio-x-generic",icon_size:80,icon_type:St.IconType.SYMBOLIC,style_class:"icon"});
        this.meta_title = new St.Label({text:_("Unknown Title"),style_class:"title"});
        this.meta_artist = new St.Label({text:_("Unknown Artist"),style_class:"artist"});
        this.meta_album = new St.Label({text:_("Unknown Album"),style_class:"album"});
        this.ctrl_prev = new St.Button({label:Prev,style_class:"control_first"});
        this.ctrl_prev.connect("clicked",Lang.bind(this,function () {
            if (this._mediaplayer_player.CanGoPrevious)
                this._mediaplayer_player.PreviousRemote();
        }));
        this.ctrl_playpause = new St.Button({label:Play,style_class:"control"});
        if (this._mediaplayer_player.PlaybackStatus=="Playing")
            this.ctrl_playpause.set_label(Pause);
        else
            this.ctrl_playpause.set_label(Play);
        this.ctrl_playpause.connect("clicked",Lang.bind(this,function () {
            if (this.ctrl_playpause.get_label()==Play && this._mediaplayer_player.CanPlay)
                this._mediaplayer_player.PlayRemote();
            else if (this.ctrl_playpause.get_label()==Pause && this._mediaplayer_player.CanPause)
                this._mediaplayer_player.PauseRemote();
        }));
        this.ctrl_stop = new St.Button({label:Stop,style_class:"control"});
        this.ctrl_stop.connect("clicked",Lang.bind(this,function () {
            if (this._mediaplayer_player.CanControl)
                this._mediaplayer_player.StopRemote();
        }));
        this.ctrl_next = new St.Button({label:Next,style_class:"control"});
        this.ctrl_next.connect("clicked",Lang.bind(this,function () {
            if (this._mediaplayer_player.CanGoNext)
                this._mediaplayer_player.NextRemote();
        }));
        this.ctrl_shuffle = new St.Button({label:ShuffleOff,style_class:"control"});
        this.ctrl_shuffle.connect("clicked",Lang.bind(this,function () {
            if (this._mediaplayer_player.CanControl && this._mediaplayer_player.Shuffle!=null)
                this._mediaplayer_player.Shuffle = !this.shuffle;
        }));
        this.ctrl_loop = new St.Button({label:LoopOff,style_class:"control_last"});
        this.ctrl_loop.connect("clicked",Lang.bind(this,function () {
            if (this._mediaplayer_player.CanControl && this._mediaplayer_player.LoopStatus!=null)
                this._mediaplayer_player.LoopStatus = this.loop=="None" ? "Track" : this.loop=="Track" ? "Playlist" : "None";
        }));
        this.info_box.add_actor(this.info_art);
        this.meta_box.add_actor(this.meta_title);
        this.meta_box.add_actor(this.meta_artist);
        this.meta_box.add_actor(this.meta_album);
        this.ctrl_box.add_actor(this.ctrl_prev);
        this.ctrl_box.add_actor(this.ctrl_playpause);
        this.ctrl_box.add_actor(this.ctrl_stop);
        this.ctrl_box.add_actor(this.ctrl_next);
        this.ctrl_box.add_actor(this.ctrl_shuffle);
        this.ctrl_box.add_actor(this.ctrl_loop);
        this.meta_box.add_actor(this.ctrl_box);
        this.info_box.add_actor(this.meta_box);
        this.info_item.addActor(this.info_box);
        this.addMenuItem(this.info_item);
        
        // Seekbar
        this.position = 0;
        this.duration = 180000000;
        this.seekbar = new PopupMenu.PopupSliderMenuItem(0.5);
        this.seekbar.connect("value-changed",Lang.bind(this,function () {
            this._mediaplayer_player.SetPositionRemote(this.trackid, Math.round(this.seekbar.value*this.duration));
        }));
        this.seekbar.removeActor(this.seekbar._slider);
        this.seek_pos = new St.Label({text:"0:00",style_class:"time"});
        this.seek_dur = new St.Label({text:"0:00",style_class:"time"});
        this.seek_table = new St.Table({homogeneous:false});
        this.seek_table.add(new St.Icon({icon_name:"document-open-recent",icon_size:16,icon_type:St.IconType.SYMBOLIC,style_class:"icon"}),{row:0,col:0,x_expand:false});
        this.seek_table.add(this.seek_pos,{row:0,col:1,x_expand:false});
        this.seek_table.add(this.seekbar._slider,{row:0,col:2,x_expand:true,x_fill:true});
        this.seek_table.add(this.seek_dur,{row:0,col:3,x_expand:false});
        this.seekbar.addActor(this.seek_table,{span:-1,align:St.Align.MIDDLE,expand:true});
        this.addMenuItem(this.seekbar);
        if (this._mediaplayer_player.Metadata) {
            mdata = this._mediaplayer_player.Metadata;
            if ("mpris:trackid" in mdata) this.trackid = mdata["mpris:trackid"].unpack();
            else this.trackid = "/org/mpris/MediaPlayer2/TrackList/NoTrack";
            if ("xesam:artist" in mdata) this.setArtist(mdata["xesam:artist"].deep_unpack()[0]);
            if ("xesam:album" in mdata) this.setAlbum(mdata["xesam:album"].unpack());
            if ("xesam:title" in mdata) this.setTitle(mdata["xesam:title"].unpack());
            if ("mpris:artUrl" in mdata) this.setArtwork(mdata["mpris:artUrl"].unpack());
            if ("mpris:length" in mdata) this.setDuration(mdata["mpris:length"].unpack());
        }
        if (this._mediaplayer_player.Position)
            this.setPosition(this._mediaplayer_player.Position);
        this._updateSeekbar();
        
        // Volume
        this.volumebar = new PopupMenu.PopupSliderMenuItem(1);
        this.volumebar.connect("value-changed",Lang.bind(this,function () {
            this._mediaplayer_player.Volume = this.volumebar.value;
        }));
        this.volumebar.removeActor(this.volumebar._slider);
        this.volume_label = new St.Label({text:"100%",style_class:"volume"});
        this.volume_icon = new St.Icon({icon_name:"audio-volume-high",icon_size:16,icon_type:St.IconType.SYMBOLIC,style_class:"icon"})
        this.volume_table = new St.Table({homogeneous:false});
        this.volume_table.add(this.volume_icon,{row:0,col:0,x_expand:false});
        this.volume_table.add(this.volumebar._slider,{row:0,col:1,x_fill:true,x_expand:true});
        this.volume_table.add(this.volume_label,{row:0,col:2,x_expand:false});
        this.volumebar.addActor(this.volume_table,{span:-1,align:St.Align.MIDDLE,expand:true});
        this.addMenuItem(this.volumebar);
        if (!this._mediaplayer_player.Volume) this.volumebar.actor.hide();
        else this.setVolume(this._mediaplayer_player.Volume);
        
        // Ratebar
        this.min_rate = 1;
        this.max_rate = 1;
        this.rate = 1;
        this.ratebar = new PopupMenu.PopupSliderMenuItem(1);
        this.ratebar.connect("value-changed",Lang.bind(this,function () {
            this._mediaplayer_player.Rate = this.ratebar.value*(this.max_rate-this.min_rate)+this.min_rate;
        }));
        this.ratebar.removeActor(this.ratebar._slider);
        this.rate_label = new St.Label({text:"100%",style_class:"rate"});
        this.rate_table = new St.Table({homogeneous:false});
        this.rate_table.add(new St.Icon({icon_name:"media-seek-forward",icon_size:16,icon_type:St.IconType.SYMBOLIC,style_class:"icon"}),{row:0,col:0,x_expand:false});
        this.rate_table.add(this.ratebar._slider,{row:0,col:1,x_expand:true,x_fill:true});
        this.rate_table.add(this.rate_label,{row:0,col:2,x_expand:false});
        this.ratebar.addActor(this.rate_table,{span:-1,align:St.Align.MIDDLE,expand:true});
        this.addMenuItem(this.ratebar);
        if (!this._mediaplayer_player.MinimumRate || !this._mediaplayer_player.MaximumRate || !this._mediaplayer_player.Rate)
            this.ratebar.actor.hide();
        else {
            this.setMinimumRate(this._mediaplayer_player.MinimumRate);
            this.setMaximumRate(this._mediaplayer_player.MaximumRate);
            this.setRate(this._mediaplayer_player.Rate);
        }
        
        if (!this._mediaplayer_player.CanControl) {
            this.ctrl_prev.set_style_pseudo_class("insensitive");
            this.ctrl_playpause.set_style_pseudo_class("insensitive");
            this.ctrl_stop.set_style_pseudo_class("insensitive");
            this.ctrl_next.set_style_pseudo_class("insensitive");
            this.ctrl_shuffle.set_style_pseudo_class("insensitive");
            this.ctrl_loop.set_style_pseudo_class("insensitive");
            this.seekbar.actor.hide();
            this.volumebar.actor.hide();
            this.ratebar.actor.hide();
        } else {
            if (!this._mediaplayer_player.CanSeek) 
                this.seekbar.actor.hide();
            if (!this._mediaplayer_player.CanGoNext)
                this.ctrl_next.set_style_pseudo_class("insensitive");
            if (!this._mediaplayer_player.CanGoPrevious)
                this.ctrl_prev.set_style_pseudo_class("insensitive");
            if (this._mediaplayer_player.Shuffle==null)
                this.ctrl_shuffle.set_style_pseudo_class("insensitive");
            else
                this.setShuffle(this._mediaplayer_player.Shuffle);
            if (this._mediaplayer_player.LoopStatus==null)
                this.ctrl_loop.set_style_pseudo_class("insensitive");
            else
                this.setLoop(this._mediaplayer_player.LoopStatus);
            pl = this._mediaplayer_player.CanPlay;
            pa = this._mediaplayer_player.CanPause;
            if (pl || pa) {
                this.ctrl_playpause.remove_style_pseudo_class("insensitive");
                if (pl && pa) {
                    if (this._mediaplayer_player.PlaybackStatus=="Playing")
                        this.ctrl_playpause.set_label(Pause);
                    else
                        this.ctrl_playpause.set_label(Play);
                } else if (pl && !pa)
                    this.ctrl_playpause.set_label(Play);
                else if (!pl && pa)
                    this.ctrl_playpause.set_label(Pause);
            } else {
                this.ctrl_playpause.set_style_pseudo_class("insensitive");
            }
        }
   
        this.plistSerial = -1;
        this._updatePlaylists(0);
        
        this.tlistSerial = -1;
        this._updateTrackList(0);
        
        this._properties.connectSignal("PropertiesChanged",Lang.bind(this,function (proxy, sender, [iface, changed]) {
            if (iface=="org.mpris.MediaPlayer2") {
                if (changed.CanQuit) {
                    if (changed.CanQuit.unpack())
                        this.quit_button.remove_style_pseudo_class("insensitive");
                    else
                        this.quit_button.set_style_pseudo_class("insensitive");
                }
                if (changed.CanRaise) {
                    if (changed.CanRaise.unpack()) 
                        this.name_item.setSensitive(true);
                    else {
                        this.name_item.setSensitive(false);
                        this.name_item.remove_style_pseudo_class("insensitive");
                    }
                }
                if (changed.CanSetFullscreen) {
                    if (changed.CanSetFullscreen.unpack())
                        this.fulls_button.remove_style_pseudo_class("insensitive");
                    else
                        this.fulls_button.set_style_pseudo_class("insensitive");
                }
                if (changed.Fullscreen) {
                    if (changed.Fullscreen.unpack())
                        this.fulls_button.set_style_pseudo_class("active");
                    else
                        this.fulls_button.remove_style_pseudo_class("active");
                }
            } else if (iface=="org.mpris.MediaPlayer2.Player") {
                if (changed.CanGoNext) {
                    if (changed.CanGoNext.unpack())
                        this.ctrl_next.remove_style_pseudo_class("insensitive");
                    else
                        this.ctrl_next.set_style_pseudo_class("insensitive");
                }
                if (changed.CanGoPrevious) {
                    if (changed.CanGoPrevious.unpack())
                        this.ctrl_prev.remove_style_pseudo_class("insensitive");
                    else
                        this.ctrl_prev.set_style_pseudo_class("insensitive");
                }
                if (changed.CanPlay || changed.CanPause) {
                    pl = changed.CanPlay ? changed.CanPlay.unpack() : this._mediaplayer_player.CanPlay;
                    pa = changed.CanPlay ? changed.CanPause.unpack() : this._mediaplayer_player.CanPause;
                    if (pl || pa) {
                        this.ctrl_playpause.remove_style_pseudo_class("insensitive");
                        if (pl && pa) {
                            if (this._mediaplayer_player.PlaybackStatus=="Playing")
                                this.ctrl_playpause.set_label(Pause);
                            else
                                this.ctrl_playpause.set_label(Play);
                        } else if (pl && !pa)
                            this.ctrl_playpause.set_label(Play);
                        else if (!pl && pa)
                            this.ctrl_playpause.set_label(Pause);
                    } else {
                        this.ctrl_playpause.set_style_pseudo_class("insensitive");
                    }
                }
                if (changed.CanSeek)
                    this._updateSeekbar();
                if (changed.CanControl) {
                    if (changed.CanControl.unpack()) {
                        if (this._mediaplayer_player.CanGoPrevious) this.ctrl_prev.remove_style_pseudo_class("insensitive");
                        this.ctrl_playpause.remove_style_pseudo_class("insensitive");
                        this.ctrl_stop.remove_style_pseudo_class("insensitive");
                        if (this._mediaplayer_player.CanGoNext) this.ctrl_next.remove_style_pseudo_class("insensitive");
                        if (this._mediaplayer_player.Shuffle) this.ctrl_shuffle.set_style_pseudo_class("insensitive");
                        if (this._mediaplayer_player.LoopStatus) this.ctrl_loop.set_style_pseudo_class("insensitive");
                        if (this._mediaplayer_player.Volume) this.volume.actor.show();
                        else this.volumebar.actor.hide();
                        this._updateSeekbar();
                        this._updateRatebar();
                    } else {
                        this.ctrl_prev.set_style_pseudo_class("insensitive");
                        this.ctrl_playpause.set_style_pseudo_class("insensitive");
                        this.ctrl_stop.set_style_pseudo_class("insensitive");
                        this.ctrl_next.set_style_pseudo_class("insensitive");
                        this.ctrl_shuffle.set_style_pseudo_class("insensitive");
                        this.ctrl_loop.set_style_pseudo_class("insensitive");
                        this.seekbar.actor.hide();
                        this.volumebar.actor.hide();
                        this.ratebar.actor.hide();
                    }
                }
                if (changed.LoopStatus) 
                    this.setLoop(changed.LoopStatus.unpack());
                if (changed.MaximumRate) 
                    this.setMaximumRate(changed.MaximumRate.unpack());
                if (changed.Metadata) {
                    mdata = changed.Metadata.deep_unpack();
                    this.trackid = "/org/mpris/MediaPlayer2/TrackList/NoTrack";
                    this.setTitle(_("Unknown Title"));
                    this.setArtist(_("Unknown Artist"));
                    this.setAlbum(_("Unknown Album"));
                    this.setArtwork("");
                    this.setDuration(0);
                    if ("mpris:trackid" in mdata) this.trackid = mdata["mpris:trackid"].unpack();
                    if ("xesam:artist" in mdata) this.setArtist(mdata["xesam:artist"].deep_unpack()[0]);
                    if ("xesam:album" in mdata) this.setAlbum(mdata["xesam:album"].unpack());
                    if ("xesam:title" in mdata) this.setTitle(mdata["xesam:title"].unpack());
                    if ("mpris:artUrl" in mdata) this.setArtwork(mdata["mpris:artUrl"].unpack());
                    if ("mpris:length" in mdata) this.setDuration(mdata["mpris:length"].unpack());
                    this._updateTrackList(this.tlistSerial+1);
                }
                if (changed.MinimumRate) 
                    this.setMinimumRate(changed.MinimumRate.unpack());
                if (changed.PlaybackStatus) {
                    if (changed.PlaybackStatus.unpack()=="Playing") {
                        if (this._mediaplayer_player.CanPause)
                            this.ctrl_playpause.set_label(Pause);
                        else
                            this.ctrl_playpause.set_label(Play);
                    }
                    else if (changed.PlaybackStatus.unpack()=="Paused") this.ctrl_playpause.set_label(Play);
                    else if (changed.PlaybackStatus.unpack()=="Stopped") {
                        this.ctrl_playpause.set_label(Play);
                        this.setTitle(_("Unknown Title"));
                        this.setArtist(_("Unknown Artist"));
                        this.setAlbum(_("Unknown Album"));
                        this.setArtwork("");
                        this.setDuration(0);
                        this._updateTrackList(this.tlistSerial+1);
                    }
                }
                if (changed.Position) 
                    this.setPosition(changed.Position.unpack());
                if (changed.Rate) 
                    this.setRate(changed.Rate.unpack());
                if (changed.Shuffle) 
                    this.setShuffle(changed.Shuffle.unpack());
                if (changed.Volume) 
                    this.setVolume(changed.Volume.unpack());
            } else if (iface=="org.mpris.MediaPlayer2.Playlists") {
                this._updatePlaylists(this.plistSerial+1);
            } else if (iface=="org.mpris.MediaPlayer2.TrackList") {
                this._updateTrackList(this.tlistSerial+1);
            }
        }));
        this._mediaplayer_player.connectSignal("Seeked",Lang.bind(this,function (proxy, sender, [pos]){
            this.setPosition(pos);
        }));
        this._mediaplayer_playlists.connectSignal("PlaylistChanged",Lang.bind(this,function () {
            this._updatePlaylists(this.plistSerial+1);
        }));
        this._mediaplayer_tracklist.connectSignal("TrackListChanged",Lang.bind(this,function () {
            this._updateTrackList(this.tlistSerial+1);
        }));
        this._mediaplayer_tracklist.connectSignal("TrackAdded",Lang.bind(this,function () {
            this._updateTrackList(this.tlistSerial+1);
        }));
        this._mediaplayer_tracklist.connectSignal("TrackRemoved",Lang.bind(this,function () {
            this._updateTrackList(this.tlistSerial+1);
        }));
        this._mediaplayer_tracklist.connectSignal("TrackMetadataCharged",Lang.bind(this,function () {
            this._updateTrackList(this.tlistSerial+1);
        }));
        this._refreshPosition();
    },
    
    _contains: function(a,obj) {
        var i = a.length;
        while (i--)
            if (a[i] === obj) 
                return true;
        return false;
    },
    
     _createTrackListItem: function(trackid,n,artist,title) {
        menuitem = new PopupMenu.PopupBaseMenuItem();
        title = title ? title : "Unknown";
        if (artist && artist!="Unknown")
            tx = ""+(parseInt(n)+1)+". "+title+" by "+artist;
        else
            tx = ""+(parseInt(n)+1)+". "+title;
        label = new St.Label({text:tx,style_class:"tracklist"});
        if (trackid==this.trackid)
            label.set_style_pseudo_class("active");
        menuitem.addActor(label);
        if (this._mediaplayer_tracklist.CanEditTracks) {
            del_button = new St.Button({label:"\u2718",style_class:"delete"});
            del_button.connect('clicked',Lang.bind(this,function () {
                this._mediaplayer_tracklist.RemoveTrackRemote(trackid);
            }));
            menuitem.addActor(del_button,{align:St.Align.END});
        }
        menuitem.connect('activate',Lang.bind(this,function () {
            this._mediaplayer_tracklist.GoToRemote(trackid);
        }));
        this.tlist_menu.menu.addMenuItem(menuitem);
    },
    
    _createPlaylistItem: function(plists,n) {
        menuitem = new PopupMenu.PopupBaseMenuItem();
        box = new St.BoxLayout();
        label = new St.Label({text:plists[0][n][1],style_class:"playlist"});
        if (this._mediaplayer_playlists.ActivePlaylist && this._mediaplayer_playlists.ActivePlaylist[0] && this._mediaplayer_playlists.ActivePlaylist[1][0]==plists[0][n][0]) {
            plicon = this._getIconForPlaylist("media-playback-start");     
            label.set_style_pseudo_class("active");       
        } else    
            plicon = this._getIconForPlaylist(plists[0][n][2]);
        box.add_actor(plicon);
        box.add_actor(label);
        menuitem.addActor(box);
        menuitem.connect('activate',Lang.bind(this,function () {
            this._mediaplayer_playlists.ActivatePlaylistRemote(plists[0][n][0]);
        }));
        this.plist_menu.menu.addMenuItem(menuitem);
    },
    
    _getIconForPlaylist: function(ico) {
        if (!ico || ico.substr(0,4)=="http") ico = "playlist"
        else if (ico.substr(0,4)=="file") {
            plicon = new St.Bin({style_class:'icon'});
            plicon.set_child(new Clutter.Texture({filename:ico.substr(7),width:16,height:16}));
            return plicon;
        }
        return new St.Icon({icon_name:ico,icon_size:16,icon_type:St.IconType.SYMBOLIC,style_class:'icon'});
    },
    
    _toTimeString: function (micros) {
        micros /= Math.floor(1000000);
        hours = Math.floor(micros/3600);
        minutes = Math.floor((micros-(hours*3600))/60);
        micros = Math.floor(micros%60);
        timestr = "0:00";
        if (hours>0) {
            timestr = ""+hours+":"+(minutes<10 ? "0"+minutes : minutes)+":"+(micros<10 ? "0"+micros : micros);
        } else {
            timestr = ""+(minutes<10 ? "0"+minutes : minutes)+":"+(micros<10 ? "0"+micros : micros);
        }
        return timestr;
    },
    
    _refreshPosition: function () {
        if (this.ctrl_playpause.get_label()==Pause)
            this.setPosition(this.position+this.rate*1000000);
        Mainloop.timeout_add_seconds(1, Lang.bind(this, this._refreshPosition));
    },
    
    _updatePlaylists: function (serial) {
        if (serial!=0)
            this.plist_menu.destroy();
        if (this._mediaplayer_playlists.PlaylistCount!=null) {
            this.plistSerial = serial;
            this.plist_menu = new PopupMenu.PopupSubMenuMenuItem("Playlists");
            this.addMenuItem(this.plist_menu);
            order = this._mediaplayer_playlists.Orderings[0];
            if (this._contains(this._mediaplayer_playlists.Orderings,"Alphabetical"))
                order = "Alphabetical";
            else if (this._contains(this._mediaplayer_playlists.Orderings,"Created"))
                order = "Created";
            this._mediaplayer_playlists.GetPlaylistsRemote(0,100,order,false,Lang.bind(this,function (plists){
                for (n in plists[0])
                    if (serial==this.plistSerial) 
                        this._createPlaylistItem(plists,n);
                    else
                        break;
            }));
        }
    },
    
    _updateRatebar: function () {
        if (this.min_rate == this.max_rate || !this._mediaplayer_player.CanControl) {
            this.ratebar.actor.hide();
        } else {
            this.ratebar.actor.show();
            this.ratebar.setValue((this.rate-this.min_rate)/(this.max_rate-this.min_rate));
        }
        rate = Math.round(this.rate*100);
        if (rate<10) rate = "  "+rate;
        else if (rate<100) rate = " "+rate;
        this.rate_label.set_text(""+rate+"%");
    },
    
    _updateSeekbar: function () {
        if (this.duration>0 && this._mediaplayer_player.CanControl && this._mediaplayer_player.CanSeek) {
            this.seekbar.actor.show();
            this.seekbar.setValue(this.position / this.duration);
        } else {
            this.seekbar.actor.hide();
        }
    },
    
    _updateTrackList: function (serial) {
        if (serial!=0)
            this.tlist_menu.destroy();
        if (this._mediaplayer.HasTrackList && this._mediaplayer_tracklist.Tracks!=null) {
            this.tlistSerial = serial;
            this.tlist_menu = new PopupMenu.PopupSubMenuMenuItem("Tracks");
            this.addMenuItem(this.tlist_menu);
            this._mediaplayer_tracklist.GetTracksMetadataRemote(this._mediaplayer_tracklist.Tracks,Lang.bind(this, function (tracks) {
                for (t in tracks[0]) {
                    if (serial!=this.tlistSerial) return;
                    this._createTrackListItem(tracks[0][t]["mpris:trackid"].unpack(),t,tracks[0][t]["xesam:artist"].deep_unpack()[0],tracks[0][t]["xesam:title"].unpack());
                }
            }));
        }
    },
    
    setAlbum: function (album) {
        this.meta_album.set_text(album);
    },
    
    setArtist: function (artist) {
        this.meta_artist.set_text(artist);
    },
    
    setArtwork: function (artwork,type) {
        this.info_box.remove_actor(this.info_art);
        this.info_box.remove_actor(this.meta_box);
        if (!artwork || artwork=="file:///usr/share/pixmaps/rsfap/noalbumart.png") artwork="audio-x-generic";
        if (artwork.substr(0,4)=="file") {
            this.info_art = new St.Bin({style_class:'icon'});
            this.info_art.set_child(new Clutter.Texture({filename:artwork.substr(7),width:80,height:80}));
        } else if (artwork.substr(0,4)=="http") {
            art = Gio.file_new_for_uri(decodeURIComponent(artwork));
            if (!this.tmpArt)
                this.tmpArt = Gio.file_new_tmp('XXXXXX.mprisplayer-artwork')[0];
            art.read_async(null, null, Lang.bind(this, function (art,result) {
                let input = cover.read_finish(result);
                let output = this.tmpArt.replace(null, false, Gio.FileCreateFlags.REPLACE_DESTINATION, null, null);
                output.splice_async(input, Gio.OutputStreamSpliceFlags.CLOSE_TARGET, 0, null, Lang.bind(this, function (){
                    outStream.splice_finish(result, null);
                    this.info_art = new St.Bin({style_class:'icon'});
                    this.info_art.set_child(new Clutter.Texture({filename:this.tmpArt.get_path(),width:80,height:80}));
                }));
            }));
        } else {
            if (!type) type=St.IconType.SYMBOLIC;
            this.info_art = new St.Icon({icon_name:artwork,icon_size:80,icon_type:type,style_class:"icon"});
        }
        this.info_box.add_actor(this.info_art);
        this.info_box.add_actor(this.meta_box);
    },
    
    setDuration: function (dur) {
        this.position = 0;
        this.duration = dur;
        this.seek_dur.set_text(this._toTimeString(dur));
        this._updateSeekbar();
    },
    
    setIcon: function (icon,type) {
        this.name_box.remove_actor(this.name_icon);
        this.name_box.remove_actor(this.name_label);
        if (!icon || icon.substr(0,4)=="http") icon="audio-x-generic";
        if (icon.substr(0,4)=="file") {
            this.name_icon = new St.Bin({style_class:'icon'});
            this.name_icon.set_child(new Clutter.Texture({filter_quality:2,filename:icon.substr(7),width:16,height:16}));
        } else {
            if (!type) type=St.IconType.SYMBOLIC;
            this.name_icon = new St.Icon({icon_name:icon,icon_size:16,icon_type:type,style_class:"icon"});
        } 
        this.name_box.add_actor(this.name_icon);
        this.name_box.add_actor(this.name_label);
    },
    
    setLoop: function (loop) {
        this.loop = loop;
        this.ctrl_loop.set_label(loop=="Playlist" ? LoopAll : loop=="Track" ? LoopTrack : LoopOff);
    },
    
    setName: function (text) {
        this.name_label.set_text(text);
    },
    
    setMaximumRate: function (mrate) {
        this.max_rate = mrate;
        this.min_rate = Math.min(this.min_rate,this.max_rate);
        this.rate = Math.max(this.min_rate,Math.min(this.max_rate,this.rate));
        this._updateRatebar();
    },
    
    setMinimumRate: function (mrate) {
        this.min_rate = Math.max(0.01,mrate);
        this.max_rate = Math.max(this.max_rate,this.min_rate);
        this.rate = Math.max(this.min_rate,Math.min(this.max_rate,this.rate));
        this._updateRatebar();
    },
    
    setPosition: function (pos) {
        this.position = pos;
        this.seek_pos.set_text(this._toTimeString(pos));
        this._updateSeekbar();
    },
    
    setRate: function (nrate) {
        this.rate = Math.max(this.min_rate,Math.min(this.max_rate,nrate));
        this._updateRatebar();
    },
    
    setShuffle: function (shuffle) {
        this.shuffle = shuffle;
        this.ctrl_shuffle.set_label(shuffle ? ShuffleOn : ShuffleOff);
    },
    
    setTitle: function (title) {
        this.meta_title.set_text(title);
    },
    
    setVolume: function (volume) {
        vol = Math.round(volume*100);
        if (vol<1) this.volume_icon.set_icon_name("audio-volume-muted");
        else if (vol<40) this.volume_icon.set_icon_name("audio-volume-low");
        else if (vol<80) this.volume_icon.set_icon_name("audio-volume-medium");
        else this.volume_icon.set_icon_name("audio-volume-high");
        if (vol<10) vol = "  "+vol;
        else if (vol<100) vol = " "+vol;
        this.volume_label.set_text(""+vol+"%");
        this.volumebar.setValue(volume);
    }
});

let dbus, sects, volind;

function add(bus,owner) {
    player = new MPRISPlayer(bus,owner);
    sepa = new PopupMenu.PopupSeparatorMenuItem();
    sepb = new PopupMenu.PopupSeparatorMenuItem();
    sects[owner] = Array(sepa,player,sepb);
    volind.menu.addMenuItem(sepa,volind.menu.length-2);
    volind.menu.addMenuItem(player,volind.menu.length-2);
    volind.menu.addMenuItem(sepb,volind.menu.length-2);
}

function change(bus,old_owner,new_owner) {
    if (old_owner in sects) {
        remove(old_owner);
        add(bus,new_owner);
    }
}

function disable() {
    for (x in sects) {
        if (sects[x][0])
            sects[x][0].actor.hide();
        if (sects[x][1])
            sects[x][1].actor.hide();
        if (sects[x][2])
            sects[x][2].actor.hide();
    }
}

function enable() {
    for (x in sects) {
        if (sects[x][0])
            sects[x][0].actor.show();
        if (sects[x][1])
            sects[x][1].actor.show();
        if (sects[x][2])
            sects[x][2].actor.show();
    }
}

function init() {
    volind = Main.panel._statusArea.volume;
    sects = {};
    sects["init"] = Array(new PopupMenu.PopupSeparatorMenuItem(),null,new PopupMenu.PopupSeparatorMenuItem());
    for (x in sects) {
        if (sects[x][0]) {
            volind.menu.addMenuItem(sects[x][0],volind.menu.length-2);
            sects[x][0].actor.hide();
        }
        if (sects[x][1]) {
            volind.menu.addMenuItem(sects[x][1],volind.menu.length-2);
            sects[x][1].actor.hide();
        }
        if (sects[x][2]) {
            volind.menu.addMenuItem(sects[x][2],volind.menu.length-2);
            sects[x][2].actor.hide();
        }
    }
    dbus = DBus.DBus();
    dbus.ListNamesRemote(Lang.bind(this,function(buses) {
            for (n in buses[0]) {
                let bus = buses[0][n];
                if (/^org\.mpris\.MediaPlayer2\./.test(bus))
                    dbus.GetNameOwnerRemote(bus, Lang.bind(this,
                        function(owner) {
                            add(bus,owner);
                        }
                    ));
           }
        }
    ));
    dbus.connectSignal("NameOwnerChanged",Lang.bind(this,function (proxy,sender,[bus,old_owner,new_owner]) {
        if (/^org\.mpris\.MediaPlayer2\./.test(bus)) {
            global.log(bus);
            global.log(old_owner);
            global.log(new_owner);
            if (!old_owner && new_owner) {
                global.log("add");
                add(bus,new_owner);
            } else if (old_owner && !new_owner) {
                global.log("remove");
                remove(old_owner);
            } else {
                global.log("changed");
                change(bus,old_owner,new_owner);
            }
            global.log("---------------");
        }
    }));
}

function remove(owner) {
    if (owner in sects) {
        if (sects[owner][0]) sects[owner][0].destroy();
        if (sects[owner][1]) sects[owner][1].destroy();
        if (sects[owner][2]) sects[owner][2].destroy();
    }
}
